@@ -1,10 +1,10 @@
 ##----------------------------------------------------------------------------
 ## A real Try Catch Block Implementation Using Perl Filter - ~/lib/Nice/Try.pm
-## Version v0.1.11
+## Version v0.2.0
 ## Copyright(c) 2021 DEGUEST Pte. Ltd.
 ## Author: Jacques Deguest <jack@deguest.jp>
 ## Created 2020/05/17
-## Modified 2021/03/26
+## Modified 2021/05/12
 ## All rights reserved
 ## 
 ## This program is free software; you can redistribute  it  and/or  modify  it
@@ -22,7 +22,7 @@ BEGIN
     use Filter::Util::Call;
     use Scalar::Util;
     use List::Util ();
-    our $VERSION = 'v0.1.11';
+    our $VERSION = 'v0.2.0';
     our $ERROR;
     our( $CATCH, $DIED, $EXCEPTION, $FINALLY, $HAS_CATCH, @RETVAL, $SENTINEL, $TRY, $WANTARRAY );
 }
@@ -343,7 +343,7 @@ sub _parse
     $self->_messagef( 3, "Results found increased from %d to %d results.", scalar( @$ref ), scalar( @$alt_ref ) );
     @$ref = @$alt_ref if( scalar( @$alt_ref ) > scalar( @$ref ) );
     
-    ## $self->_message( 3, "Script code is now:\n'$elem'" );
+    # $self->_message( 3, "Script code is now:\n'$elem'" );
     
     foreach my $this ( @$ref )
     {
@@ -577,7 +577,9 @@ EOT
             {
                 $try_def->{block} = $emb;
             }
-                        
+            
+            $self->_process_loop_breaks( $try_def->{block} );
+            
             ## my $try_block = $try_def->{block}->content;
             my $try_block = $self->_serialize( $try_def->{block} );
             $try_block =~ s/^\{[[:blank:]]*|[[:blank:]]*\}$//gs;
@@ -588,7 +590,7 @@ CORE::local \$Nice::Try::TRY = CORE::sub
     CORE::do __TRY_OPEN_NL__{ __BLOCK_PLACEHOLDER__ };__TRY__CLOSE_NL__
     CORE::return( \$Nice::Try::SENTINEL );
 };
-CORE::local ( \$Nice::Try::EXCEPTION, \$Nice::Try::DIED, \@Nice::Try::RETVAL );
+CORE::local ( \$Nice::Try::EXCEPTION, \$Nice::Try::DIED, \@Nice::Try::RETVAL, \@Nice::Try::VOID );
 __FINALLY_BLOCK__ CORE::local \$Nice::Try::HAS_CATCH = $has_catch_clause;
 CORE::local \$Nice::Try::WANTARRAY = CORE::wantarray;
 {
@@ -605,7 +607,7 @@ CORE::local \$Nice::Try::WANTARRAY = CORE::wantarray;
         }
         else 
         {
-            &\$Nice::Try::TRY;
+            \@Nice::Try::VOID = &\$Nice::Try::TRY;
         }
     };
     \$Nice::Try::DIED = CORE::length( \$\@ ) ? 1 : 0;
@@ -724,6 +726,7 @@ EOT
                 }
                 ## $self->_message( 3, "\$i = $i, \$total_catch = $total_catch and cond = '$cond'" );
                 ## my $block = $cdef->{block}->content;
+                $self->_process_loop_breaks( $cdef->{block} );
                 my $block = $self->_serialize( $cdef->{block} );
                 $block =~ s/^\{[[:blank:]]*|[[:blank:]]*\}$//gs;
                 my $catch_section = '';
@@ -744,7 +747,7 @@ EOT
             } 
             else 
             {
-                \&\$Nice::Try::CATCH;
+                \@Nice::Try::VOID = \&\$Nice::Try::CATCH;
             }
 EOT
                 if( $cdef->{var} )
@@ -882,13 +885,65 @@ if( CORE::defined( \$Nice::Try::WANTARRAY ) and
       ( Scalar::Util::blessed( \$Nice::Try::RETVAL[0] ) && !\$Nice::Try::RETVAL[0]->isa( 'Nice::Try::SENTINEL' ) ) 
     ) ) 
 {
+    if( \$Nice::Try::RETVAL[0] eq '__NEXT__' )
+    {
+        \$Nice::Try::BREAK = 'next';
+    }
+    elsif( \$Nice::Try::RETVAL[0] eq '__LAST__' )
+    {
+        \$Nice::Try::BREAK = 'last';
+    }
+    elsif( \$Nice::Try::RETVAL[0] eq '__REDO__' )
+    {
+        \$Nice::Try::BREAK = 'redo';
+    }
     CORE::return( \$Nice::Try::WANTARRAY ? \@Nice::Try::RETVAL : \$Nice::Try::RETVAL[0] );
 }
+elsif( scalar( \@Nice::Try::VOID ) )
+{
+    if( \$Nice::Try::VOID[0] eq '__NEXT__' )
+    {
+        \$Nice::Try::BREAK = 'next';
+    }
+    elsif( \$Nice::Try::VOID[0] eq '__LAST__' )
+    {
+        \$Nice::Try::BREAK = 'last';
+    }
+    elsif( \$Nice::Try::VOID[0] eq '__REDO__' )
+    {
+        \$Nice::Try::BREAK = 'redo';
+    }
+}
 EOT
         $last_return_block =~ s/\n/ /gs unless( $self->{debug_code} );
         push( @$repl, $last_return_block );
         my $try_catch_code = join( '', @$repl );
-        my $token = PPI::Token->new( "; \{ $try_catch_code \}" ) || die( "Unable to create token" );
+        # my $token = PPI::Token->new( "; \{ $try_catch_code \}" ) || die( "Unable to create token" );
+        # XXX 2021-05-11 (Jacques): Need to remove blocks so that next or last statements can be effective.
+        my $envelop = <<EOT;
+; CORE::local \$Nice::Try::BREAK;
+\{
+__TRY_CATCH_CODE__
+\}
+if( \$Nice::Try::BREAK )
+{
+    if( \$Nice::Try::BREAK eq 'next' )
+    {
+        CORE::next;
+    }
+    elsif( \$Nice::Try::BREAK eq 'last' )
+    {
+        CORE::last;
+    }
+    elsif( \$Nice::Try::BREAK eq 'redo' )
+    {
+        CORE::redo;
+    }
+}
+EOT
+        $envelop =~ s/\n/ /gs unless( $self->{debug_code} );
+        $envelop =~ s/__TRY_CATCH_CODE__/$try_catch_code/;
+        my $token = PPI::Token->new( $envelop ) || die( "Unable to create token" );
         $token->set_class( 'Structure' );
         ## $self->_messagef( 3, "Token is '$token' and of class '%s' and inherit from PPI::Token? %s", $token->class, ($token->isa( 'PPI::Token' ) ? 'yes' : 'no' ) );
         my $struct = PPI::Structure->new( $token ) || die( "Unable to create PPI::Structure element" );
@@ -917,6 +972,81 @@ EOT
     return( $elem );
 }
 
+sub _process_loop_breaks
+{
+    my $self = shift( @_ );
+	my $elem = shift( @_ ) || return( '' );
+    return( $elem ) if( !$elem->children );
+    $self->_messagef( 5, "Checking %d elements for '$elem'", scalar( $elem->elements ) );
+    foreach my $e ( $elem->elements )
+    {
+        $self->_messagef( 6, "Checking element: [%d] class %s with %d children and value '%s'\n", $e->line_number, $e->class, ( $e->can('elements') ? scalar( $e->elements ) : 0 ), $e->content );
+        my $class = $e->class;
+        # We found a for, foreach or while loops and we skip, because if there are any break words (next, last, redo) inside, it is not our problem.
+        if( $class eq 'PPI::Structure::For' ||
+            ( $class eq 'PPI::Statement::Compound' && $e->first_element->content =~ /^(foreach|while)$/ ) )
+        {
+            # $self->_message( 6, "Skipping it. Its first word was '", $e->first_element->content, "'" );
+            next;
+        }
+        elsif( $class eq 'PPI::Statement::Break' )
+        {
+            my $words = $e->find( 'PPI::Token::Word' );
+            $self->_messagef( 5, "Found %d word elements inside break element.", scalar( @$words ) );
+            $self->_message( 5, "Word 1 -> ", $words->[0]->content );
+            $self->_message( 5, "Word 2 -> ", $words->[1]->content ) if( scalar( @$words ) > 1 );
+            # $self->_browse( $e );
+            # If we found a break word without a label, i.e. next, last, redo, 
+            # we replace it with a special return statement
+            if( ( scalar( @$words ) == 1 ||
+                  ( scalar( @$words ) > 1 && $words->[1]->content =~ /^(for|foreach|given|if|unless|until|while)$/ )
+                ) && 
+                ( $words->[0]->content eq 'next' ||
+                  $words->[0]->content eq 'last' ||
+                  $words->[0]->content eq 'redo' ) )
+            {
+                # We add our special return value. Notice that we use 'return' and not 
+                # 'CORE::return'. See below why.
+                my $break_code = qq{return( '__} . uc( $words->[0]->content ) . qq{__' )};
+                # e.g. next if( $i == 2 );
+                # next and if are both treated as 'word' by PPI
+                if( scalar( @$words ) > 1 )
+                {
+                    ( my $ct = $e->content ) =~ s/^(next|last|redo)//;
+                    $break_code .= $ct;
+                }
+                else
+                {
+                    $break_code .= ';'
+                }
+                $self->_message( 5, "Replacing this node with: $break_code" );
+                my $break_doc = PPI::Document->new( \$break_code, readonly => 1 );
+                my $new_elem = $break_doc->first_element;
+                # $self->_browse( $new_elem );
+                $new_elem->remove;
+                $self->_message( 5, "New element is object '", sub{ overload::StrVal( $new_elem ) }, "' -> $new_elem" );
+                # Not yet implemented as of 2021-05-11 dixit PPI, so we use a hack to make it available anyhow
+                $e->replace( $new_elem );
+                # 2021-05-12 (Jacques): I have to do this workaround, because weirdly enough
+                # PPI (at least with PPI::Node version 1.270) will refuse to add our element
+                # if the 'return' word is 'CORE::return' so, we add it without and change it after
+                $new_elem->first_element->set_content( 'CORE::return' );
+                # $self->_message( 5, "return litteral value is: ", $new_elem->first_element->content );
+            }
+            next;
+        }
+        
+        if( $e->can('elements') && $e->elements )
+        {
+            $self->_process_loop_breaks( $e );
+        }
+    }
+    # $self->_message( 5, "Element now is: '", sub{ $elem->content }, "'" );
+    $self->_message( 5, "Element now is: '$elem'" );
+    # $self->_browse( $elem );
+    return( $elem );
+}
+
 ## Taken from PPI::Document
 sub _serialize 
 {
@@ -1092,6 +1222,21 @@ sub _serialize
     }
 }
 
+{
+    package
+        PPI::Element;
+    
+    sub replace {
+        my $self    = ref $_[0] ? shift : return undef;
+        # If our object and the other are not of the same class, PPI refuses to replace 
+        # to avoid damages to perl code
+        my $other = _INSTANCE(shift, ref $self) or return undef;
+        # die "The ->replace method has not yet been implemented";
+        $self->parent->__replace_child( $self, $other );
+        1;
+    }
+}
+
 1;
 
 __END__
@@ -1119,7 +1264,7 @@ Nice::Try - A real Try Catch Block Implementation Using Perl Filter
     }
     # Some comment
     catch( Exception $e ) {
-        return( "Caught an exception \$e" );
+        return( "Caught an exception $e" );
     }
     # More comment with space too
 
@@ -1143,7 +1288,7 @@ When run, this would produce, as one would expect:
 
 =head1 VERSION
 
-    v0.1.11
+    v0.2.0
 
 =head1 DESCRIPTION
 
@@ -1242,6 +1387,45 @@ Would produce:
     I am trying my best Jacques!
     Failed: But I failed
 
+=item * C<try> or C<catch> blocks can contain flow control keywords such as C<next>, C<last> and C<redo>
+
+    while( defined( my $product = $items->[++$i] ) )
+    {
+        try
+        {
+            # Do something
+            last if( !$product->active );
+        }
+        catch( $oops )
+        {
+            log->( "Error: $oops" );
+            last;
+        }
+    }
+    continue
+    {
+        try
+        {
+            if( $product->region eq 'Asia' )
+            {
+                push( @asia, $product );
+            }
+            else
+            {
+                next;
+            }
+        }
+        catch( $e )
+        {
+            $log->( "An unexpected error has occurred. Is $product an object? $e" );
+            last;
+        }
+    }
+
+=item * Can be used with or without a C<catch> block
+
+=item * Supports a C<finally> block called in void context for cleanup for example
+
 =back
 
 =head1 WHY USE IT?
@@ -1324,7 +1508,7 @@ In group 3, L<TryCatch> was working wonderfully, but was relying on L<Devel::Dec
 
 In group 4, there is L<Syntax::Keyword::Try>, which is a great alternative if you do not care about exception variable assignment or exception class filter. You can only use C<$@>
 
-Since L<perl version 5.33|https://perldoc.perl.org/blead/perlsyn#Try-Catch-Exception-Handling> you can use the try-catch block using an experimental feature which may be removed in future versions, by writing:
+Since L<perl version 5.33.7|https://perldoc.perl.org/blead/perlsyn#Try-Catch-Exception-Handling> you can use the try-catch block using an experimental feature which may be removed in future versions, by writing:
 
     use feature 'try'; # will emit a warning this is experimental
 
@@ -1351,7 +1535,7 @@ But B<you cannot do>:
     {
         return( $self->error( "Something went awry with MyException: $oh_well" ) );
     }
-    # No support for 'finally' yet in perl 5.34
+    # No support for 'finally' yet in perl version 5.33.7
     finally
     {
         # do some cleanup here
@@ -1389,6 +1573,21 @@ However, because this is designed for clean-up, it is called in void context, so
 
 L<Nice::Try> can be used with a single C<try> block which will, in effect, behaves like an eval and the special variable C<$@> will be available as always.
 
+    try
+    {
+        die( "Oh no, something went wrong!\n" );
+    }
+    print( "Got here with $@\n" );
+
+or even:
+
+    try
+    {
+        die( "Oh no, something went wrong!\n" );
+    }
+    catch( $e ); # Not very meaningful, but it will work
+    print( "Got here with $@\n" );
+
 However, if you decide to catch class exceptions, make sure to add a default C<catch( $e )>. For example:
 
     try
@@ -1448,6 +1647,84 @@ Since, try-catch block can be nested, the following would work too:
         # do something about it
     }
 
+=head1 LOOPS
+
+Since version v0.2.0 L<Nice::Try> supports the use of flow control keywords such as C<next>, C<last> and C<redo> inside try-catch blocks. For example:
+
+    my @names = qw( John Jack Peter Paul Mark );
+    for( $i..$#names )
+    {
+        try
+        {
+            next if( $i == 2 );
+            # some more code...
+        }
+        catch( $e )
+        {
+            print( "Got exception: $e\n" );
+        }
+    }
+
+It also works inside the catch block or inside the C<continue> block:
+
+    while( defined( my $product = $items->[++$i] ) )
+    {
+        # Do something
+    }
+    continue
+    {
+        try
+        {
+            if( $product->region eq 'Asia' )
+            {
+                push( @asia, $product );
+            }
+            else
+            {
+                next;
+            }
+        }
+        catch( $e )
+        {
+            $log->( "An unexpected error has occurred. Is $product an object? $e" );
+            last;
+        }
+    }
+
+Control flow with labels also work
+
+    ELEM: foreach my $n ( @names )
+    {
+        try
+        {
+            $n->moveAfter( $this );
+            next ELEM if( $n->value == 1234567 );
+        }
+        catch( $oops )
+        {
+            last ELEM;
+        }
+    }
+
+However, if you enclose a try-catch block inside another block, use of C<next>, C<last> or C<redo> will silently not work. This is due to perl control flow. See L<perlsyn> for more information on this. For example, the following would not yield the desired outcome:
+
+    ELEM: foreach my $n ( @names )
+    {
+        { # <--- Here is the culprit
+            try
+            {
+                $n->moveAfter( $this );
+                # This next statement will not do anything.
+                next ELEM if( $n->value == 1234567 );
+            }
+            catch( $oops )
+            {
+                # Neither would this one.
+                last ELEM;
+            }
+        }
+    }
+
 =head1 DEBUGGING
 
 And to have L<Nice::Try> save the filtered code to a file, pass it the C<debug_file> parameter like this:
@@ -1480,7 +1757,7 @@ Jacques Deguest E<lt>F<jack@deguest.jp>E<gt>
 
 =head1 SEE ALSO
 
-L<PPI>, L<Filter::Util::Call>, L<Try::Harder>, L<Syntax::Keyword::Try>
+L<PPI>, L<Filter::Util::Call>, L<Try::Harder>, L<Syntax::Keyword::Try>, L<Exception::Class>
 
 =head1 COPYRIGHT & LICENSE
 
